@@ -33,7 +33,8 @@
 (declare-function agent-shell--indent-string "agent-shell")
 (declare-function agent-shell--make-acp-client "agent-shell")
 (declare-function agent-shell-make-agent-config "agent-shell")
-(declare-function agent-shell-start "agent-shell")
+(autoload 'agent-shell-make-agent-config "agent-shell")
+(declare-function agent-shell--dwim "agent-shell")
 
 (cl-defun agent-shell-openai-make-authentication (&key api-key codex-api-key login)
   "Create OpenAI authentication configuration.
@@ -111,6 +112,7 @@ Returns an agent configuration alist using `agent-shell-make-agent-config'."
   (when (and (boundp 'agent-shell-openai-key) agent-shell-openai-key)
     (user-error "Please migrate to use agent-shell-openai-authentication and eval (setq agent-shell-openai-key nil)"))
   (agent-shell-make-agent-config
+   :identifier 'codex
    :mode-line-name "Codex"
    :buffer-name "Codex"
    :shell-prompt "Codex> "
@@ -132,8 +134,8 @@ Returns an agent configuration alist using `agent-shell-make-agent-config'."
 (defun agent-shell-openai-start-codex ()
   "Start an interactive Codex agent shell."
   (interactive)
-  (agent-shell-start
-   :config (agent-shell-openai-make-codex-config)))
+  (agent-shell--dwim :config (agent-shell-openai-make-codex-config)
+                     :new-shell t))
 
 (cl-defun agent-shell-openai-make-codex-client (&key buffer)
   "Create a Codex client using configured authentication with BUFFER as context.
